#!/usr/bin/perl
#$Id: CGI-remote-server.pl,v 1.14 2013/04/17 07:39:03 sqko Exp $
use strict;
use warnings;
use IO::Socket;
use IO::Interface::Simple;
use SemaforeFile;
use ElchemeaHeader;
use Elchemea;
use ElchemeaUser;
use ElchemeaProgram;
use Sys::Hostname;
my $VERSION = 0.3;
my $PROGRAM = 'CGI-remote-server.pl';
my $debug = 0;

if (scalar(@ARGV) > 0)
{
    for(my $x = 0; $x < scalar(@ARGV);$x++)
    {
	if ($ARGV[$x] =~ /--version/)
	{
	    print "$PROGRAM is: $VERSION\n";
	    exit 0;
	}
	if ($ARGV[$x] =~ /--debug/)
	{
	    $debug = 1;
	}
    } 
}


#### Try to get ip address of netwrok interface ###
my @ilst = ('eth','em','lo');
my $local = '';
my @ifl = IO::Interface::Simple->interfaces();
OUTER:
foreach my $try (@ilst)
{
    foreach my $interface (@ifl)
    {
        if ($interface =~ /$try/)
        {
            my $if = IO::Interface::Simple->new($interface);
            $local = $if->address();
            last OUTER if ($local);
        }
    }
}
### Use localhost if everything else fails ###
$local = 'localhost' unless ($local);

our $host = $local;
our $port    =  get_config_value('remote','port');
our $user      = "nobody";
print "Using host $host and port $port\n" if ($debug);
# Let children perish
#
$SIG{"CHLD"} = "IGNORE";

# Get userid for $user. Abort if userid is zero (superuser) or
# non-existent.
#
unless ( our $uid = ( getpwnam($user) )[2] ) 
{
    die "Attempt to run server as non-existent or superuser\n";
}

# Initialize C structure#
our $server_addr   = ( gethostbyname($host) )[4];
our $server_struct = pack( "S n a4 x8", AF_INET, $port, $server_addr );
print "Server address: $server_addr\n" if ($debug);

# Set up socket#
our $proto = ( getprotobyname("tcp") )[2];
socket( SOCK, PF_INET, SOCK_STREAM, $proto )
    || die "Failed to initialize socket: $!\n";

# Bind to address/port and set up pending queue #
setsockopt( SOCK, SOL_SOCKET, SO_REUSEADDR, 1 )  || die "setsockopt() failed: $!\n";
bind( SOCK, $server_struct ) || die "bind() failed: $!\n";
listen( SOCK, SOMAXCONN ) || die "listen() failed: $!\n";

our %cmdlst = ();
$cmdlst{'debug'} = \&debug;
$cmdlst{'exit'} = \&shutdown;
$cmdlst{'impedance'} = \&impedance;
$cmdlst{'chrono'} = \&chrono;
$cmdlst{'potsweep'} = \&potsweep;
$cmdlst{'get_file'} = \&get_file;
$cmdlst{'get_file_now'} = \&get_file_now;
$cmdlst{'ping'} = \&ping;
$cmdlst{'mode'} = \&mode;
$cmdlst{'session'} = \&session;
$cmdlst{'reset_potentiostat'} = \&reset_potentiostat;
$cmdlst{'change_channel'} = \&change_channel;



# Deal with requests #
our $dot_addr;
our $remote_host;
for ( ; ; ) {

    # Grab next pending request
    #
    $remote_host = accept( NEWSOCK, SOCK );
    die "accept() error: $!\n" unless ($remote_host);
    ### Process request ###
    &request();
    close(NEWSOCK);
}

# Close master socket
close(SOCK);
exit;

sub request
{
    # Resolve hostname of remote machine before calling chroot()
    our $raw_addr = ( unpack( "S n a4 x8", $remote_host ) )[2];
    $dot_addr = join ( ".", unpack( "C4", $raw_addr ) );
    our $name = ( gethostbyaddr( $raw_addr, AF_INET ) )[0];
    # Read client request and get $path
    our @args = ();
    our $cmd = undef;
    my $ltime = localtime;
    print "$ltime\t$dot_addr\t$name\t" if ($debug);
    my $str = <NEWSOCK>;
    chomp $str;
    ($cmd,@args) = split(/[\s\t]+/,$str);
    if ($cmdlst{$cmd})
    {
	my $ok = '';
	eval
	{
	    $ok = $cmdlst{$cmd}->(@args);
	};
	if ($@)
	{
	    warn($@);
	}
        print "'$cmd'\tResult: $ok\n" if ($debug);
        print NEWSOCK $ok;
    }
    else
    {
        print "Unknown command: '$cmd'\n";
        print NEWSOCK -1;
    }
}

sub session
{
    my $user = ElchemeaUser->new(shift);
    return $user->session();
}

sub impedance
{
    my $user = ElchemeaUser->new(shift);
    my $mode = shift;
    my $session = shift;
    $user->proglog("Impedance started remotely from $dot_addr");
    return Elchemea::CGI_client('impedance',$user->name,$session);
}

sub chrono
{
    my $user = ElchemeaUser->new(shift);
    my $mode = shift;
    my $session = shift;
    $user->proglog("Chrono started remotely from $dot_addr");
    return Elchemea::CGI_client('chrono',$user->name,$session);
}

sub potsweep
{
    my $user = ElchemeaUser->new(shift);
    my $mode = shift;
    my $session = shift;
    $user->proglog("Potsweep started remotely from $dot_addr");
    return Elchemea::CGI_client('potsweep',$user->name,$session);
}

sub mode
{
    my $compat = get_config_value('global','compatibility');
    if ($compat && $compat =~ /yes/i)
    {
	return '1260';
    }
    else
    {
	my $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address'));
	my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
	my $str = $pot->type.':'.$fra->type;
	$str =~ s/\s+/_/g;
	return $str;
    }
}

sub ping
{
    my $hostname = hostname;
    return "Elchemea system on $hostname on addr:$host listening on port $port";
}

sub change_channel
{
    my $id = ElchemeaProgram->new(shift);
    my $mode = shift; ### Dummy ###
    my $chid = shift;
    $id->change_channel($chid);
    return "Changed multiplexer channel to $chid";
}

sub reset_potentiostat
{
    my $id = ElchemeaProgram->new(shift);
    $id->reset_potentiostat;
    return "Potentiostat reset";
}

sub get_file
{
    return 'MEASURING' if (Elchemea::is_locked);
    return get_file_now(@_);
}

sub get_file_now
{
    my $user = ElchemeaUser->new(shift);
    my $mode = shift;
    my $session = shift;
    my $fileid = shift;
    $fileid = sprintf("%04d",$fileid) if ($fileid < 1000);
    my $file = $userdir.$user->name.'/'.$session.'/s'.$session.'_'.$fileid.'.';
    my $filename = '';
    foreach ('i2b','i2c','i2p')
    {
	my $f = $file.$_;
	next unless -e $f;
	$filename = $file. $_;
	last;
    }
    return '' unless $filename;
    my $id = SemaforeFile->new($filename);
    my $string = join("\n",$id->readlines);
    return $string;
}

sub debug
{
    $debug = !$debug;
    return "Debug: $debug";
}

sub shutdown
{
    print "\n\nShutting down server!\n";
    print NEWSOCK "Shutting down server!";
    close(SOCK);
    exit;
}
