#!/usr/bin/perl
#$Id: gpib_to_serial_server.pl,v 1.2 2014/12/12 10:30:44 sofc Exp $
use POSIX qw(strftime);
use Device::SerialPort;
use Fcntl qw(:DEFAULT :flock);
use IO::Socket;
use strict;
use warnings;
use Sys::Hostname;

my $debug = 0;
if (!@ARGV)
{
    die "Not enough parameters for $0\nUsage: $0 ALIASDEVICE\nWhere the ALIASDEVICE is the TTY (default ttyS0) which the server connects to.\nThe server listens on localhost port 12345 thus emulating a true PGIB-server by redirecting the requests to the serial server.\n\n";
}

##### Setup hash of functions ####
my %cmdlst = ();
$cmdlst{'q'} = \&quit;
$cmdlst{'d'} = \&set_debug;
$cmdlst{'w'} = \&writestanford;
$cmdlst{'r'} = \&readback;

#### Base port number is 202020, added to this is the ttys nr ###
#### EG ttyS0 => 202020, ttyS4 => 202024 etc.                 ###
my $port = 202020;

$| = 1;


my $pre = sprintf "%c%c%c%c%c", 255, 255, 255, 255, 255;
my $cmdpre = sprintf "%c%c", 130, 138;

my $buffer = '';

our $remote = shift;
our $remote_port = shift;
$remote_port += $port;

### Bind to socket ###
my $sock = new IO::Socket::INET(
    LocalHost => 'localhost',
    LocalPort => 12345,
    Proto  => 'tcp',
    Listen => 10,
    Reuse  => 1
);
die "Could not create socket: $!\n" unless $sock;
$| = 1;
my $new_sock;
while ( $new_sock = $sock->accept() ) 
{
    if ($debug)
    {
	print strftime "%Y:%m:%d:%H:%M:%S", localtime;
    }
    my $other_end = getpeername($new_sock);
    my ( $port, $iaddr ) = unpack_sockaddr_in($other_end);
    my $ip_address = inet_ntoa($iaddr);
    print "  ",$ip_address, "\n" if ($debug);
    if (my $str = <$new_sock>) {
        my ($cmd,@args) = split (/[\s\t]+/,$str);
	$cmd = lc($cmd);
	print $str if ($debug);
	if ($cmd && $cmdlst{$cmd})
	{
	    my $res = $cmdlst{$cmd}->(@args);
	    print "Result: '$res'\n\n" if ($debug);
	    print $new_sock $res if (defined($res));
	}
	close $new_sock;
    }
    else
    {
	close $new_sock;
    }
}
close($sock);
exit;

sub readback
{
    my $val = $buffer;
    $buffer = undef;
    return $val;
}

sub writestanford
{
    my $addr = shift;
    $buffer = &remote_connect('icp_raw',@_);
}

sub quit
{
    print $new_sock "quit command received, exiting", "\n\n";
    close $new_sock;
    close $sock;
    exit;
}

sub set_debug
{
    my $cmd = shift;
    if ($cmd)
    { 
	if ($cmd =~ /enable/i)
	{
	    $debug = 1;
	    return "Debug on";
	}
	if ($cmd =~ /disable/i)
	{
	    $debug = 0;
	    return "Debug off";
	}
    }
    else
    {
	if ($debug)
	{
	    $debug = 0;
	    return "Debug off";
	}
	$debug = 1;
	return "Debug on";
    }
}


sub remote_connect(@)
{
    my ($iaddr, $paddr, $proto, $line);
    if ($remote_port =~ /\D/) { $remote_port = getservbyname($remote_port, 'tcp') }
    die "No port" unless $remote_port;
    $iaddr   = inet_aton($remote)               || die "no host: $remote";
    $paddr   = sockaddr_in($remote_port, $iaddr);
    my $out = '';
    $proto   = getprotobyname('tcp');
    eval {
        socket(SOCK, PF_INET, SOCK_STREAM, $proto)  || die "socket: $!";
        connect(SOCK, $paddr)    || die "Could not connect to $remote port $remote_port: $!
";
        select(SOCK);
        $|=1;
        select(STDOUT);
        print SOCK join("\t",@_),"\n\n";
        print SOCK "\n\n";
        while (defined($line = <SOCK>))
        {
            $out = $out.$line;
        }
        close (SOCK)            || die "close: $!";
    };
    if ($@)
    {
        print $@;
        return undef;
    }
    $out =~ s/[\r\n]*//g;
    return $out;
}
