#!/usr/bin/perl -wT
#$Id: run_chrono.pl,v 1.47 2012/10/02 11:48:26 sofc Exp $
$ENV{'PATH'} = '/bin:/usr/bin:';
use strict;
use warnings;
use Time::HiRes qw(sleep time);
use ElchemeaUser;
use Elchemea;
use ElchemeaProgram qw();
use IO::Handle;
use IO::AtomicFile;
use Sys::Hostname;
use File::Copy;
use Graphics::GnuplotIF qw(GnuplotIF);
$SIG{INT} = \&my_sigint;

unless (@ARGV > 1)
{
    die "Not enough parameters for $0. Usage: '$0 user_name session_nr [opt polarisation] [opt display]'";
}
## To do: get gnuplot plotting working ###
my $plot = 0;
my $user = ElchemeaUser->new(shift);
my $debug = Elchemea::get_config_value('global','debug');
Elchemea::debug($debug);
$user->debug($debug);
my $session = shift;
my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
my $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address'));

my $fork = undef;
$fork = shift if (@ARGV);

#### Get setup ####
my %pots = ();
foreach my $tag ($pot->setup_tags_chrono)
{
    $pots{$tag} = $user->get_config_value('chrono',$tag,$session);
}
foreach my $tag ('totaltime','totaltime_unit','sleeptime','keep_potential','mode','polarisation_offset_program')
{
    $pots{$tag} = $user->get_config_value('chrono',$tag,$session);
}
$pots{'totaltime'} = 1 unless ($pots{'totaltime'});
$pots{'totaltime'} = $pots{'totaltime'} * 60 if ($pots{'totaltime_unit'} && $pots{'totaltime_unit'} =~ /m/);
$pots{'totaltime'} = $pots{'totaltime'} * 60 * 60 if ($pots{'totaltime_unit'} && $pots{'totaltime_unit'} =~ /h/);

$pots{'totaltime'} = $fork if (defined($fork) && $fork =~ /\d+/);
if (@ARGV)
{
    ### Use explicit polarisation if third argument is defined ###
    $pots{'polarisation'} = shift;
}
#### Reverse polarisation due to working electrode on ground terminal ###
$pots{'polarisation'} = -$pots{'polarisation'} if (exists($pots{'polarisation'}) && defined($pots{'polarisation'}) && $pots{'polarisation'} !~ /ocv/i);
$pots{'polarisation_offset'} = Elchemea::get_external_polarisation_function($pots{'polarisation_offset_program'}) if (exists($pots{'polarisation_offset_program'}));

my $measurelock;
unless ($pots{'totaltime'} == 0)
{
    $measurelock = Elchemea::lock_measure();
    unless ($measurelock)
    {
	my $string = "Exiting $0: someone else is measuring!";
	Elchemea::EFAlog($string);
	warn($string."\n");
	exit;
    }
}
else
{
    $pot->set_ignore();
    $measurelock = $ElchemeaHeader::chrono_runforever_semafore;
    open MYOUT,'>',$measurelock or die "Could not open file '$measurelock' for writing $!";
    chmod 0666,$measurelock;
    close MYOUT;
}

my $modedesc = '';
$modedesc = 'amperometry' if ($pots{'mode'} =~ /Volt/i);
$modedesc = 'potentiometry' if ($pots{'mode'} =~ /Amp/i);
my $totaltime = $pots{'totaltime'};
Elchemea::set_proglog((time+$totaltime),'Initialising chrono '.$modedesc,$user->name);
sleep(0.3); ### Allow other programs to finish and clean up ###
my $nr = $user->increase_measure_nr;
$user->writedebug("Done fetching configuration. Chrono $modedesc nr $nr started");
$user->proglog("Chrono $modedesc nr $nr started");
#### Generate data file header strings ###
my $hostname = hostname;
my $identifier_string = "Chrono $modedesc nr $nr for user ".$user->name." session ".$user->session." on '$hostname'\nUsing potentiostat '".$pot->type."' and frequency response analyser '".$fra->type."'";
my $test_id = 'Desc: '.$user->sample_name;
my $logstr = $user->log_string;
$test_id .= " \# ".$logstr if ($logstr);
my $loctime = time;
my $date_string = join(':',Elchemea::get_time('now')). ' '. $loctime;
my $mode = $pot->mode;
my $pol = -$pots{'polarisation'};
$mode = '' unless (defined($mode));
$pol = '' unless (defined($pol));
my $genmodestr = "Bias: $pol $mode";
#### Setup gnuplot display ###
my $gpid = undef;
$gpid = Graphics::GnuplotIF->new( style=>'lines') if (join(' ',@ARGV) =~ /display/i);
#### Setup measuring system ###
$user->writedebug("Setting up measuring systems");
my $logfile = $user->get_measure_file('.log');
my $shortfile = $user->get_measure_file('.i2c');
&plot();
###  Get termperatue and voltages at start if defined ###
my $tempstart = '';
$tempstart = $user->get_channels() if ($user->get_config_value('measure','measure_temp') && $user->get_config_value('measure','measure_temp') =~ /start/i);

$pot->set_debugstream($logfile);
$pot->debugon;
unless ($fra->isallive)
{
    my $str = "Error: Frequency response analyser not responding!\n";
    $pot->writedebug($str);
    $user->proglog($str);
    $user->proglog('Chrono terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    die $str;
}
unless ($pot->isallive)
{
    my $str = "Error: Potentiostat device not responding!\n";
    $pot->writedebug($str);
    $user->proglog($str);
    $user->proglog('Chrono terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    die $str;
}
$fra->reset;
$pot->writedebug('Saving original setup from devices');
my @potsave = $pot->save;
$fra->writedebug('Initialising devices for chrono '.$modedesc);
$pot->writedebug($tempstart) if ($tempstart);
$pot->set_mode($pots{'mode'});
my $truepol = -$pot->setup_chrono(\%pots);
$genmodestr .= ', Polarisation offset (external): '.(-$pot->last_pol_offset).' '.$pot->mode;
$genmodestr .= ', True polarisation (Fixed): '.$truepol.' '.$pot->mode;
my $message = Elchemea::get_last_message;
open MYOUT,'>',$shortfile or die "Could not open file $shortfile for writing $!";
print MYOUT <<END;
$hostname:$shortfile
$test_id
$date_string
$genmodestr
$message
$tempstart
END

    close MYOUT;


$pot->writedebug('Running chrono '.$modedesc);
$user->writedebug('Starting chrono '.$modedesc);
$pot->start_chrono if (exists($pots{'polarisation'}) && $pots{'polarisation'} =~ /\d/); ### Only engage source if not in ocv ###
my $endtime = time + $pots{'totaltime'};
my $starttime = time;
my $sleeptime = $pots{'sleeptime'};
sleep($pots{'delay'}) if ($pots{'delay'});
Elchemea::set_proglog((time+$pots{'totaltime'}),'Running chrono '.$modedesc,$user->name);
my $plottime = 3;
my $pt = time + $plottime;
OUTER:
while (time < $endtime || $pots{'totaltime'} == 0)
{
    my $st = $sleeptime;
    $st = 1 if ($st > 1);
    if ($pots{'totaltime'} == 0)
    {
	&bfile(&get_dvm) unless Elchemea::is_locked_device;
	last OUTER unless -e $measurelock;
	my $et = time + $sleeptime;
	while (time < $et)
	{
	    sleep($st);
	    last OUTER unless -e $measurelock;
	}
    }
    else
    {
	&bfile(&get_dvm);
	last OUTER unless Elchemea::is_locked;
	my $et = time + $sleeptime;
	while (time < $et)
	{
	    last OUTER unless Elchemea::is_locked;
	    sleep($st);
	}
	next OUTER if (time < $pt || $pots{'totaltime'} == 0);
	&plot;
	$pt = time + $plottime;
    }
}
my $error = $pot->get_error();
my $code = $pot->errorcode($error);
$error = $code if ($code);
$user->proglog("Errors encountered durring chrono: ".$error) unless (!$error || $error =~ /^[0\s]+$/);
$user->proglog("Chrono $modedesc stopped") unless (-e $measurelock || Elchemea::is_locked);
unless (exists($pots{'keep_potential'}) && defined($pots{'keep_potential'}) && $pots{'keep_potential'} =~ /yes/i)
{
    $fra->writedebug('Restoring original device configuration');
    $pot->bulk_load_check(@potsave);
}
my $tend = '';
$tend = $user->get_channels() if ($user->get_config_value('measure','measure_temp') && $user->get_config_value('measure','measure_temp') =~ /end/i); 
&bfile($tend) if ($tend);
$pot->writedebug($tend) if ($tend);
sleep(0.3);
Elchemea::set_proglog(time,'Finished',$user->name) unless ($pots{'totaltime'} == 0);
Elchemea::unlock_measure unless ($pots{'totaltime'} == 0);
$user->writedebug('Done running chrono '.$modedesc);
exit unless ($gpid);
$gpid->gnuplot_cmd("set title 'Finished'");
$gpid->gnuplot_cmd("rep");
print "Press return to exit.\n";
while (my $test = <STDIN>)
{
    last if ($test =~ /^[\s\t\n]*$/);
}
exit;



sub plot()
{
    my $plotcmd = "set yrange[0:1]\nf(x) = 0\nplot f(x) linetype -1";
    $plotcmd = "plot \"<awk 'NR>4&&NF>1' $shortfile\" u (\$1-$starttime):2 tit 'Potential' with lines,'' u (\$1-$starttime):4 axes x1y2 tit 'Current' with lines" if ($shortfile && -e $shortfile);
    #### Setup ragnes here (in case they change) ###
    $starttime = time unless ($shortfile && -e $shortfile); ### Guard for initial empty plot ###
    my $xr = '';
    $xr = 60 if (time < $starttime + 60);
    unless ($gpid)
    {
        my $tmpfile = $Elchemea::plotfile.'.tmp';
        ### initialise gnuplot and run in sub to ensure close (file not updated otherwise) ###
        {
            my $gp = Graphics::GnuplotIF->new( style=>'lines');
            $gp->gnuplot_cmd('set nokey');
	    $gp->gnuplot_cmd('set zeroaxis linetype -1');
	    $gp->gnuplot_cmd('set x2zeroaxis linetype -1');
	    $gp->gnuplot_cmd("set ylabel 'Potential (Volt)'");
	    $gp->gnuplot_cmd("set y2label 'Current (A)'");
	    $gp->gnuplot_cmd("set xla 'time (seconds)'"); 
	    $gp->gnuplot_cmd("set xrange[0:$xr]");
            $gp->gnuplot_cmd("set term ".Elchemea::get_config_value('GNUPLOT','term'));
            $gp->gnuplot_cmd("set out '$tmpfile'");
            $gp->gnuplot_cmd('set size '.get_config_value('GNUPLOT','chronosize'));
            $gp->gnuplot_cmd('set y2tics');
            $gp->gnuplot_cmd('set ytics nomirror');
            $gp->gnuplot_cmd($plotcmd);
        }
        move($tmpfile,$Elchemea::plotfile);
        return undef;
    }
    #### Gnuplot commands for command line display ####
    #### Setup ragnes here (in case they change) ###
    if ($plot)
    {
	$gpid->gnuplot_cmd("set xrange[0:$xr]");
        $gpid->gnuplot_cmd("rep");
    }
    else
    {
        ### Setup all gnuplot settings here ####
	$gpid->gnuplot_cmd('set nokey');
        $gpid->gnuplot_cmd('set zeroaxis linetype -1');
	$gpid->gnuplot_cmd('set x2zeroaxis linetype -1');
	$gpid->gnuplot_cmd("set ylabel 'Potential (Volt)'");
	$gpid->gnuplot_cmd("set y2label 'Current (A)'");
	$gpid->gnuplot_cmd("set xla 'time'"); 
	$gpid->gnuplot_cmd("set xrange[0:$xr]");
	$gpid->gnuplot_cmd('set y2tics');
	$gpid->gnuplot_cmd('set ytics nomirror');
        $gpid->gnuplot_cmd($plotcmd);
        $plot = 1;
    }
    return undef;
}

sub get_dvm()
{
    my $res = $pot->run_dvm;
    my $time = time;
    return '' unless $res;
    my @r = split(/[\s\t]+/,$res);
    ### Note this is only true for default setup of the potentiostat! ###
    foreach (@r)
    {
	$_ = -$_;
    }
    return "$time $r[0] V, $r[1] A";
}


sub bfile($)
{
    my $string = shift;
    chomp $string;
    open MYOUT,'>>',$shortfile or die "Could not open file $shortfile for appending $!";
    print MYOUT $string,"\n";
    close MYOUT;
}



sub my_sigint
{
    ### Exits gracefully when catching a sigint (kill -2)!! ###
    $SIG{INT} = 'IGNORE';  ## Ignores further SIGINT ###
    #### reset potentiostat and frequency response analyser to initial state ###
    #### Do it fast with no check (we have recieved a INT signal after all)  ###
    $pot->bulk_load(@potsave);
    $pot->unset_ignore();
    Elchemea::unlock_measure;
}
