#!/usr/bin/perl -wT
#$Id: run_imp.pl,v 1.94 2014/10/24 09:44:49 sofc Exp $
$ENV{'PATH'} = '/bin:/usr/bin:';
use strict;
use warnings;
use Time::HiRes qw(sleep time);
use Elchemea;
use ElchemeaUser;
use IO::Handle;
use IO::AtomicFile;
use Sys::Hostname;
use File::Copy;
use Graphics::GnuplotIF qw(GnuplotIF);

$SIG{INT} = \&my_sigint;

unless (@ARGV > 1)
{
    die "Not enough parameters for $0. Usage: '$0 user_name session_nr [opt display]'";
}
## To do: get gnuplot plotting working ###
my $plot = 0;
my $user = ElchemeaUser->new(shift);
my $debug = Elchemea::get_config_value('global','debug');
Elchemea::debug($debug);
$user->debug($debug);
my $session = shift;
my $modtime = $user->modtime($session);
my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
my $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address'));

#### Get setup ####
my %pots = ();
my %fras = ();
foreach my $tag ($pot->setup_tags)
{
    $pots{$tag} = $user->get_config_value('impedance_potentiostat',$tag,$session);
}
foreach my $tag ($fra->setup_tags)
{
    $fras{$tag} = $user->get_config_value('impedance_FRA',$tag,$session);
}
foreach my $tag ('mode','polarisation_offset_program')
{
    $pots{$tag} = $user->get_config_value('impedance_potentiostat',$tag,$session);
    $fras{$tag} = $user->get_config_value('impedance_FRA',$tag,$session);
}
### Reverse polarisations due to convention of working electrode being gound ###
$fras{'polarisation'} = -$fras{'polarisation'} if (exists($fras{'polarisation'}) && defined($fras{'polarisation'}));
$pots{'polarisation'} = -$pots{'polarisation'} if (exists($pots{'polarisation'}) && defined($pots{'polarisation'}));
### Get external polarisation functions ###
$fras{'polarisation_offset'} = Elchemea::get_external_polarisation_function($fras{'polarisation_offset_program'}) if (exists($fras{'polarisation_offset_program'}));
$pots{'polarisation_offset'} = Elchemea::get_external_polarisation_function($pots{'polarisation_offset_program'}) if (exists($pots{'polarisation_offset_program'}));
#####
my $source = $fra->get_default('source_mode');
$source = $fras{'source_mode'} if (exists($fras{'source_mode'}) && defined($fras{'source_mode'}) && $fras{'source_mode'});
my @frqlst = $user->get_impedance_frequencies($session,$pot,$fra);
my $totaltime = $fra->get_impedance_time(\@frqlst) + $fra->setuptime + $pot->setuptime;
my $measurelock = Elchemea::lock_measure(); 
unless ($measurelock)
{
    my $string = "Exiting $0: someone else is measuring!";
    Elchemea::EFAlog($string);
    warn($string."\n");
    exit;
}
$fras{'semafore_file'} = $measurelock;
Elchemea::set_proglog((time+$totaltime),'Initialising impedance',$user->name,$source);
sleep(0.3); ## Allow other programs to finish and clean up before proceeding ###
my $nr = $user->increase_measure_nr;
$user->writedebug("Done loading configuration. Impedance nr $nr started");
$user->proglog("Impedance nr $nr started");
#### Generate data file header strings ###
my $hostname = hostname;
my $identifier_string = "Impedance nr $nr for user ".$user->name." session ".$user->session." on '$hostname'\nUsing potentiostat '".$pot->type."' and frequency response analyser '".$fra->type."'";
my $test_id = 'Desc: '.$user->sample_name;
my $logstr = $user->log_string;
$test_id .= " \# ".$logstr if ($logstr);
my $loctime = time;
my $date_string = join(':',Elchemea::get_time('now')). ' '. $loctime;
my $mode = $pots{'mode'};
my $pol = undef;
$pol = -$pots{'polarisation'} if (exists($pots{'polarisation'}) && defined($pots{'polarisation'}));
my $amplitude = $fras{'amplitude'};
if ($pot->virtual)
{
    $mode = $fras{'mode'};
    $pol = $fras{'polarisation'};
}
else
{
    ### Force DC polarisation of FRA to 0 if a potentiostat is used ### 
    $fras{'polarisation'} = 0;
}
$mode = '' unless (defined($mode));
$pol = '' unless (defined($pol));
$amplitude = '' unless ($amplitude);
my $genmodestr = "Bias: $pol $mode, Ampltude $amplitude $mode";
my $expnp = scalar(@frqlst);
$genmodestr .= ", Expected $expnp points";
my $tempstring = Elchemea::get_multiplex_info;
#### Setup gnuplot display ###
my $gpid = undef;
$gpid = Graphics::GnuplotIF->new( style=>'points') if (@ARGV);
#### Setup measuring system ###
my $logfile = $user->get_measure_file('.log');
my $shortfile = $user->get_measure_file('.i2b');
my $longfile = $user->get_measure_file('_long.i2b');
&plot();
###  Get termperatue and voltages at start if defined ###
my $tempstart = '';
$tempstart = $user->get_channels() if ($user->get_config_value('measure','measure_temp') && $user->get_config_value('measure','measure_temp') =~ /start/i);
$pot->set_debugstream($logfile);
$fra->set_debugstream($logfile);
$fra->debugon;
unless ($fra->isallive)
{
    my $str = "Error: Frequency response analyser not responding!\n";
    $fra->writedebug($str);
    $user->proglog($str);
    $user->proglog('Impedance terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    die $str;
}
unless ($pot->isallive)
{
    my $str = "Error: Potentiostat device not responding!\n";
    $fra->writedebug($str);
    $user->proglog($str);
    $user->proglog('Impedance terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    die $str;
}
$pot->debugon unless ($pot->virtual);
$fra->writedebug('Saving original setup from devices');
my @potsave = $pot->save;
my @frasave = $fra->save;
$fra->writedebug($tempstart) if ($tempstart);
$user->writedebug('Initialising devices for impedance');
$fra->writedebug('Initialising devices for impedance');
my @freqlst_out = ();
foreach my $i (@frqlst)
{
    my @l = ();
    foreach my $key (sort keys %{$i})
    {
	push @l , $key.': '.$$i{$key};
    }
    push @freqlst_out, join("  ",@l);
}
$fra->writedebug("Frequency list:\n".join("\n",@freqlst_out));
my @messages = ();
my $truepol;
eval {
    unless ($pot->virtual)
    {
	my $res = $fra->init();
	unless ($res)
	{
	    $user->writedebug("FRA did not return a valid identifier from init() Exiting!");
	    Elchemea::unlock_measure;
	    exit;
	}
	$fra->set_mode('Volt');
	$pot->set_mode($pots{'mode'});
    }
    else
    {
	$fra->set_mode($fras{'mode'});
    }
    $truepol = -$pot->setup_impedance(\%pots);
};
if ($@)
{
    $user->proglog('Error: '.$@);
    $user->proglog('Impedance terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    exit;
}
push @messages,Elchemea::get_last_message if (Elchemea::get_last_message);
Elchemea::unset_message();
### Make sure that user gets the emplitude he/she wants ###
$fras{'amplitude'} = $fras{'amplitude'} / $pot->AC_gain if (defined($fras{'amplitude'})); 
my $truepolfra = -$fra->setup_impedance(\%fras);
push @messages,Elchemea::get_last_message if (Elchemea::get_last_message);
Elchemea::unset_message();
$fras{'function'} = \&plot;
$tempstring .= ', ' if ($tempstring);
foreach my $m (@messages)
{
    chomp $m;
    $m =~ s/[\n\r]+/\, /g
}
$tempstring .= join(', ',@messages) if (@messages);
if ($tempstart)
{
    $tempstring .= ', ' if ($tempstring);
    $tempstring .= 'Start: '.$tempstart;
}
if ($pot->virtual)
{
    sleep($fra->delay());
    $genmodestr .= ', Polarisation offset: '.(-$fra->last_pol_offset);
    $genmodestr .= ', True polarisation (Fixed): '.$truepolfra.' '.$fra->mode;
    $genmodestr .= ', Delay before impedance: '.$fra->delay().' s';
}
else
{
    sleep($pot->delay());
    $genmodestr .= ', Polarisation offset (external): '.(-$pot->last_pol_offset).' '.$pot->mode;
    $genmodestr .= ', True polarisation (Fixed): '.$truepol.' '.$pot->mode;
    if ($truepolfra)
    {
	$genmodestr .= ', Polarisation (FRA): '.$fras{'polarisation'}.' '.$fra->mode;
	$genmodestr .= ', Polarisation offset (FRA): '.(-$fra->last_pol_offset).' '.$fra->mode;
	$genmodestr .= ', True polarisation (FRA, Fixed):'.$truepolfra.' '.$fra->mode;
    }
    if ($tempstring)
    {
	$tempstring .= ', Measured start polarisation: '.&get_dvm;
    }
    else
    {
	$tempstring .= 'Measured start polarisation: '.&get_dvm;
    }
    $genmodestr .= ', Delay before impedance: '.$pot->delay().' s';
}
$genmodestr .= ', Delay before each frequency: '.$fra->delay().' s';
$genmodestr .= ', Source mode: '.$source;
$fra->writedebug(join("\n",'',$identifier_string,$test_id,$date_string,$genmodestr,$tempstring,'',''));
my @datashort = ();
my @datalong = ();
my $datastring = '0.0000 0.0000 0.0000';
my $maxp = scalar(@frqlst);
my $count = 0;
$user->writedebug('Starting impedance scan');
### Run impedance scan ###
my $waserror = 0;
my $waswarn = 0;
OUTER:
while (@frqlst)
{
    my $ttime = $fra->get_impedance_time(\@frqlst);
    my $error = $fra->errorcode($fra->last_error());
    my $warning = $fra->warningcode($fra->last_warning());
    ### Unset warnings and errors for initial write (as no error or warning can be set yet) ###
    $error = '' unless ($count);  
    $warning = '' unless ($count); 
    if ($error =~ /\w/)
    {
	$waserror = 1;
    }
    if ($warning =~/\w/)
    {
	$waswarn = 1;
    }
    my $remaining = $maxp - $count;
    Elchemea::set_proglog((time+$ttime),'Running impedance',$user->name,$source,$datastring,$error,$warning,$remaining,Elchemea::get_last_message);
    my $item = shift(@frqlst);
    foreach my $tag ($pot->frequency_segment_tags)
    {
	$pot->tag_value($tag,$$item{$tag});
    }
    foreach my $tag ($fra->frequency_segment_tags)
    {
	$fra->tag_value($tag,$$item{$tag});
    }
    unless ($pot->virtual)
    {
	$fras{'resistor'} = $pot->get_resistor();
	if ($pot->mode =~ /A/i)
	{
	    ### Set ac amplitude to match resistor value ###
	    ### As fra is always in voltage mode         ###
	    $fra->amplitude($fras{'amplitude'}*$fras{'resistor'});
	}
    }
    my $res = undef;
    my $x = 3;
    while (!defined($res) && 0 < $x--)
    {
	Elchemea::set_proglog((time+$ttime),'Running impedance',$user->name,$source,$datastring,$error,$warning,$remaining,Elchemea::get_last_message);
	$res = $fra->run_frequency_all($$item{'frequency'},$$item{'time'},\%fras);
    }
    last unless Elchemea::is_locked;
    $res = $$item{'frequency'}.' 0.0000 0.0000' unless (defined($res) || length($res) < 2);
    push @datalong,$res;
    my @l = split(/\s*\#\s*/,$res);
    push @datashort,$l[0];
    $datastring = $l[0];
    &bfile($longfile,@datalong);
    &bfile($shortfile,@datashort);
    ### Force plot ###
    $modtime = 0; 
    &plot();
    $count++;
}
$user->proglog("Impedance stopped") unless (-e $measurelock);
$user->proglog("Errors were encountred, see filelog for details") if ($waserror);
$user->proglog("Warnings were encountred, see filelog for details") if ($waswarn);
unless ($pot->virtual)
{
    my $error = $pot->get_error();
    my $code = $pot->errorcode($error);
    $error = $code if ($code);
    $user->proglog("Errors encountered by potentiostat durring impedance: ".$error) unless (!$error || $error =~ /^[0\s]+$/);
}
#### reset potentiostat and frequency response analyser to initial state ###
$fra->reset;
$tempstring .= ', Measured end polarisation: '.&get_dvm unless ($pot->virtual);
@datalong = $fra->get_data_long;
@datashort = $fra->get_data;
my $tempend = '';
$tempend = $user->get_channels() if ($user->get_config_value('measure','measure_temp') && $user->get_config_value('measure','measure_temp') =~ /end/i);
if ($tempend)
{
    $fra->writedebug($tempend);
    $tempstring .= ', ' if ($tempstring);
    $tempstring .= 'End: '.$tempend;
}
&bfile($longfile,@datalong);
&bfile($shortfile,@datashort);
$fra->writedebug('Restoring original device configuration');
$fra->bulk_load_check(@frasave);
$pot->bulk_load_check(@potsave);
sleep(0.3); 
Elchemea::set_proglog(time,'Finished',$user->name);
Elchemea::unlock_measure;
$user->writedebug('Done running impedance');

exit unless ($gpid);
$gpid->gnuplot_cmd("set title 'Finished'");
$gpid->gnuplot_cmd("rep");
print "Press return to exit.\n";
while (my $test = <STDIN>)
{
    last if ($test =~ /^[\s\t\n]*$/);
}
exit;

sub get_dvm()
{
    my $res = $pot->run_dvm;
    return '' unless $res;
    my @r = split(/[\s\t]+/,$res);
    ### Note this is only true for default setup of the potentiostat! ###
    foreach (@r)
    {
	### Reverse polarisations ###
	$_ = -$_;
    }
    return "$r[0] V, $r[1] A";
}


sub bfile($)
{
    my $file = shift;
    my @data = @_;
    my $np = scalar(@data);
    my $fh = IO::AtomicFile->new( $file, ">" ) or die "Can't open $file : $!";
    print $fh <<END;
$hostname:$file
$test_id
$date_string
$genmodestr
$tempstring

$np
END

    foreach my $s (@data)
    {
	$s = '' unless defined($s);
        print {$fh} "$s\n";
    }
    $fh->close;
}


sub plot()
{
    my $plotcmd = "f(x) = 0\nplot f(x) linetype -1";
    $plotcmd = "plot \"<awk 'NR>5&&NF>1' $shortfile\" u (log(\$1)/log(10)):(\$3<0?-\$3:1/0) pt 3,'' u (log(\$1)/log(10)):(\$3>0?\$3:1/0) pt 3,'' u (log(\$1)/log(10)):(\$2>0?\$2:1/0) pt 2,'' u (log(\$1)/log(10)):(\$2<0?-\$2:1/0) pt 2" if ($shortfile && -e $shortfile);
    #### Setup ragnes here (in case they change) ###
    my $minf = log($user->get_config_value('gnuplot','bode_plot_min_freq'))/log(10);
    my $maxf = log($user->get_config_value('gnuplot','bode_plot_max_freq'))/log(10);
    my $minz = $user->get_config_value('gnuplot','bode_plot_min_imp')+0;
    my $maxz = $user->get_config_value('gnuplot','bode_plot_max_imp')+0;
    unless ($gpid)
    {
	if ($modtime != $user->modtime($session))
	{
	    $modtime = $user->modtime($session);
	    my $tmpfile = $Elchemea::plotfile.'.tmp';
	    ### initialise gnuplot and run in sub to ensure close (file not updated otherwise) ###
	    {	
		my $gp = Graphics::GnuplotIF->new( style=>'points');
		$gp->gnuplot_cmd('set nokey');
		$gp->gnuplot_cmd("set xrange [$minf:$maxf]");
		$gp->gnuplot_cmd("set yrange [$minz:$maxz]");
		$gp->gnuplot_cmd("set term ".Elchemea::get_config_value('GNUPLOT','term'));
		$gp->gnuplot_cmd("set out '$tmpfile'");
		$gp->gnuplot_cmd('set size '.get_config_value('GNUPLOT','impsize'));
		$gp->gnuplot_cmd('set logscale y');
		$gp->gnuplot_cmd("set yla 'Z (Ohm)'");
		$gp->gnuplot_cmd("set xla 'log(frequency (Hz))'");
		$gp->gnuplot_cmd($plotcmd);
	    }
	    move($tmpfile,$Elchemea::plotfile);
	}
	return undef;
    }
    #### Gnuplot commands for command line display ####
    #### Setup ragnes here (in case they change) ###
    $gpid->gnuplot_cmd("set xrange [$minf:$maxf]");
    $gpid->gnuplot_cmd("set yrange [$minz:$maxz]");
    if ($plot)
    {
	$gpid->gnuplot_cmd("rep");
    }
    else
    {
	### Setup all gnuplot settings here ####
	$gpid->gnuplot_cmd('set nokey');
	$gpid->gnuplot_cmd('set logscale y');
	$gpid->gnuplot_cmd("set yla 'Z (Ohm)'");
	$gpid->gnuplot_cmd("set xla 'log(frequency (Hz))'");
	$gpid->gnuplot_cmd($plotcmd);
	$plot = 1;
    }
    return undef;
}


sub my_sigint
{
    ### Exits gracefully when catching a sigint (kill -2)!! ###
    $SIG{INT} = 'IGNORE';  ## Ignores further SIGINT ###
    #### reset potentiostat and frequency response analyser to initial state ###
    #### Do it fast with no check (we have recieved a INT signal after all)  ###
    $fra->reset;
    $fra->bulk_load(@frasave);
    $pot->bulk_load(@potsave);
    Elchemea::unlock_measure;
}
