#!/usr/bin/perl -wT
#$Id: run_square.pl,v 1.5 2011/01/20 14:16:57 sofc Exp $

$ENV{'PATH'} = '/bin:/usr/bin:';
use strict;
use warnings;
use Time::HiRes qw(sleep time);
use Elchemea;
use ElchemeaUser;
use ElchemeaProgram qw();
use IO::Handle;
use IO::AtomicFile;
use Sys::Hostname;
use File::Copy;
use Graphics::GnuplotIF qw(GnuplotIF);
$SIG{INT} = \&my_sigint;
unless (@ARGV > 1)
{
    die "Not enough parameters for $0. Usage: '$0 user_name session_nr polarisation [opt display]'";
}
my $plot = 0;
my $user = ElchemeaUser->new(shift);
my $session = shift;
my $debug = Elchemea::get_config_value('global','debug');
Elchemea::debug($debug);
$user->debug($debug);
my $pot = Elchemea::new_PSTAT(Elchemea::get_config_value('equipment','potentiostat'),Elchemea::get_config_value('equipment','potentiostat_address'));
my $fra = Elchemea::new_FRA(Elchemea::get_config_value('equipment','FRA'),Elchemea::get_config_value('equipment','FRA_address'));
#### Get setup ####
my %pots = ();
my %fras = ();
foreach my $tag ($pot->setup_tags_chrono,'polarisation_offset_program')
{
    $pots{$tag} = $user->get_config_value('chrono',$tag,$session);
}

my $pol = shift;
### Get external polarisation functions ###
$pots{'polarisation_offset'} = Elchemea::get_external_polarisation_function($pots{'polarisation_offset_program'}) if (exists($pots{'polarisation_offset_program'}));
#####

my $measurelock = Elchemea::lock_measure();
unless ($measurelock)
{
    my $string = "Exiting $0: someone else is measuring!";
    Elchemea::EFAlog($string);
    warn($string."\n");
    exit;
}
$pots{'semafore_file'} = $measurelock;


my $modedesc = '';
$modedesc = 'Square wave amperometry' if ($pots{'mode'} =~ /Volt/i);
$modedesc = 'Square wave voltametry' if ($pots{'mode'} =~ /Amp/i);

my $nr = $user->increase_measure_nr;
my $ttime = $pot->square_time(\%pots);
Elchemea::set_proglog((time+$ttime),'Initialising '.lc($modedesc),$user->name);
sleep(0.3); ## Allow other programs to finish and clean up before proceeding ##
$user->writedebug("Done getting configuration. Square wawe output nr $nr started");
$user->proglog("Square wawe output nr $nr started with polarisation $pol for $ttime s");
#### Generate data file header strings ###
my $hostname = hostname;
my $identifier_string = "Square wawe nr $nr for user ".$user->name." session ".$user->session." on '$hostname'\nUsing potentiostat '".$pot->type."' and frequency response analyser '".$fra->type."'";
my $test_id = 'Desc: '.$user->sample_name;
my $logstr = $user->log_string;
$test_id .= " \# ".$logstr if ($logstr);
my $loctime = time;
my $date_string = join(':',Elchemea::get_time('now')). ' '. $loctime;
my $mode = $pot->mode;
$mode = '' unless (defined($mode));
#### Setup gnuplot display ###
my $gpid = undef;
$gpid = Graphics::GnuplotIF->new( style=>'points') if (@ARGV);
#### Setup measuring system ###
$user->writedebug("Setting up measurement systems");
my $logfile = $user->get_measure_file('.log');
my $shortfile = $user->get_measure_file('.i2p');
&plot();
###  Get termperatue and voltages at start if defined ###
my $tempstart = '';
$tempstart = $user->get_channels() if ($user->get_config_value('measure','measure_temp') && $user->get_config_value('measure','measure_temp') =~ /start/i);

$pot->set_debugstream($logfile);
$fra->set_debugstream($logfile);
$pot->debugon;
unless ($pot->isallive)
{
    my $str = "Error: Potentiostat device not responding!\n";
    $pot->writedebug($str);
    $user->proglog($str);
    $user->proglog('Square wave output terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    die $str;
}
unless ($fra->isallive)
{
    my $str = "Error: Frequency response analyser not responding!\n";
    $pot->writedebug($str);
    $user->proglog($str);
    $user->proglog('Square wave output terminated');
    Elchemea::set_proglog(time,'Finished',$user->name);
    Elchemea::unlock_measure;
    die $str;
}
$fra->debugon unless ($fra->virtual);
$pot->writedebug('Saving original setup from devices');
my @potsave = $pot->save;
$pot->writedebug($tempstart) if ($tempstart);
$pot->writedebug('Initialising devices for '.lc($modedesc));
$user->writedebug('Initialising devices for '.lc($modedesc));
$fra->reset;
$pot->set_mode($pots{'mode'});
$pots{'reference_potential'} = 'Fixed' unless ($pots{'reference_potential'});
my ($truepol,$totaltime) = $pot->setup_square(-$pol,\%pots);
my $genmodestr = 'Mode: '. $pot->mode.', True polarisation (first point): '.$truepol.' '.$pot->mode;
$genmodestr .= ', Polarisation offset (external): '.(-$pot->last_pol_offset).' '.$pot->mode;
$genmodestr .= ',Reference potential: '.$pots{'reference_potential'};
my $message = Elchemea::get_last_message; 
$pot->writedebug($genmodestr);
$pot->writedebug($message) if ($message);
#open MYOUT,'>',$shortfile or die "Could not open file $shortfile for writing $!";
#print MYOUT <<END;
#$hostname:$shortfile
#$test_id
#$date_string
#$genmodestr
#$tempstart
#END
#
#    close MYOUT;

$pot->writedebug('Starting '.lc($modedesc));
$user->writedebug('Starting '.lc($modedesc));
my $starttime = time;
$pot->start_square;
my $endtime = time + $totaltime;
Elchemea::set_proglog($endtime,'Running '.lc($modedesc),$user->name);

 OUTER:
    while (time < $endtime)
{
    sleep(1);
    last OUTER unless Elchemea::is_locked;
}
my $error = $pot->get_error();
my $code = $pot->errorcode($error);
$error = $code if ($code);
$user->proglog("Errors encountered durring square wave output: ".$error) unless (!$error || $error =~ /^[0\s]+$/);
$user->proglog("Potential sweep stopped!") unless (-e $measurelock);
unless (exists($pots{'keep_potential'}) && defined($pots{'keep_potential'}) && $pots{'keep_potential'} =~ /yes/i)
{
    $fra->writedebug('Restoring original device configuration');
    $pot->bulk_load_check(@potsave);
}
my $tend = '';
$tend = $user->get_channels() if ($user->get_config_value('measure','measure_temp') && $user->get_config_value('measure','measure_temp') =~ /end/i);
#&bfile($tend) if ($tend);
$pot->writedebug($tend) if ($tend);
sleep(0.3);
Elchemea::set_proglog(time,'Finished',$user->name);
Elchemea::unlock_measure;
$user->writedebug('Done running '.lc($modedesc));
exit unless ($gpid);

$gpid->gnuplot_cmd("set title 'Finished'");
$gpid->gnuplot_cmd("rep");
print "Press return to exit.\n";
while (my $test = <STDIN>)
{
    last if ($test =~ /^[\s\t\n]*$/);
}
exit;


sub plot()
{
    my $plotcmd = "set yrange[-1:1]\nf(x) = 0\nplot f(x) linetype -1";
    $plotcmd = "plot \"<awk 'NR>4&&NF>1' $shortfile\" u 2:4 tit '' with lines"  if ($shortfile && -e $shortfile);
    #### Setup ragnes here (in case they change) ###
    $starttime = time unless ($shortfile && -e $shortfile); ### Guard for initial empty plot ###
    my $xr = '';
    $xr = 60 if (time < $starttime + 60);
    unless ($gpid)
    {
        my $tmpfile = $Elchemea::plotfile.'.tmp';
        ### initialise gnuplot and run in sub to ensure close (file not updated otherwise) ###
        {
            my $gp = Graphics::GnuplotIF->new( style=>'lines');
            $gp->gnuplot_cmd('set nokey');
	    $gp->gnuplot_cmd("set title 'No graph avaliable for square wawe'");
            $gp->gnuplot_cmd('set zeroaxis linetype -1');
            $gp->gnuplot_cmd("set xlabel 'Potential (Volt)'");
            $gp->gnuplot_cmd("set ylabel 'Current (A)'");
            $gp->gnuplot_cmd("set term ".Elchemea::get_config_value('GNUPLOT','term'));
            $gp->gnuplot_cmd("set out '$tmpfile'");
            $gp->gnuplot_cmd('set size '.get_config_value('GNUPLOT','chronosize'));
            $gp->gnuplot_cmd($plotcmd);
        }
        move($tmpfile,$Elchemea::plotfile);
        return undef;
    }
    #### Gnuplot commands for command line display ####
    #### Setup ragnes here (in case they change) ###
    if ($plot)
    {
        $gpid->gnuplot_cmd("rep");
    }
    else
    {
        ### Setup all gnuplot settings here ####
        $gpid->gnuplot_cmd('set nokey');
	$gpid->gnuplot_cmd("set title 'No graph avaliable for square wawe'");
	$gpid->gnuplot_cmd('set zeroaxis linetype -1');
        $gpid->gnuplot_cmd("set xlabel 'Potential (Volt)'");
        $gpid->gnuplot_cmd("set ylabel 'Current (A)'");
        $gpid->gnuplot_cmd($plotcmd);
        $plot = 1;
    }
    return undef;
}

sub get_dvm()
{
    my $res = $pot->run_dvm;
    my $time = time;
    return '' unless $res;
    my @r = split(/[\s\t]+/,$res);
    ### Note this is only true for default setup of the potentiostat! ###
    foreach (@r)
    {
        $_ = -$_;
    }
    return "$time $r[0] V, $r[1] A";
}


sub bfile($)
{
    my $string = shift;
    chomp $string;
    open MYOUT,'>>',$shortfile or die "Could not open file $shortfile for appending $!";
    print MYOUT $string,"\n";
    close MYOUT;
}


sub my_sigint
{
    ### Exits gracefully when catching a sigint (kill -2)!! ###
    $SIG{INT} = 'IGNORE';  ## Ignores further SIGINT ###
    #### reset potentiostat and frequency response analyser to initial state ###
    #### Do it fast with no check (we have recieved a INT signal after all)  ###
    $pot->bulk_load(@potsave);
    Elchemea::unlock_measure;
}

